;-------------------------------------------------------------------------
;------------------------------- Annotation ------------------------------
;-------------------------------------------------------------------------

; ⟨...⟩        sequence formation
; s ↓ k        kth member of the sequence s (1-based) 
; #s           length of sequence s
; s§t          concatenation of sequences s and t
; s†k          drop the first k members of sequence s
; t → a, b     McCarthy conditional “if t then a else b”
; ρ[x/i]       substitution “ρ with x for i”
; x in D       injection of x into domain D
; x|D          projection of x to domain D

;-------------------------------------------------------------------------
;---------------------------- Abstract syntax ----------------------------
;-------------------------------------------------------------------------

; K ∈ Con        constants
; I ∈ Ide        variables
; E ∈ Exp        expressions
; Γ ∈ Com = Exp  commands
;
; E → K | I | (E0 E*)
;       | (begin E+)
;       | (define I E)
;       | (define (I0 I*) E+)
;       | (define (I0 I+ . I) E+)
;       | (define (I0 . I) E+)
;       | (if E0 E1) | (if E0 E1 E2)
;       | (lambda (I*) E+)
;       | (lambda (I+ . I) E+)
;       | (lambda I E+)
;       | (quote E)
;       | (set! I E)

;-------------------------------------------------------------------------
;----------------------------Domain equations-----------------------------
;-------------------------------------------------------------------------

; α ∈ L locations
; ν ∈ N natural numbers
; T = { false, true } booleans
; Q symbols
; H characters
; R numbers
; Ep = L × L pairs
; Ev = L* vectors
; Es = H* strings
; M = { false, true, null, unspecified } miscellaneous
; φ ∈ F = L × (E* → K → C) procedure values
; ε ∈ E = Q + H + R + Ep + Ev + Es + M + F expressed values
; σ ∈ S ⊂ L × E stores
; ρ ∈ U ⊂ Ide × L environments    
; κ ∈ K = E → K expression continuations

;-------------------------------------------------------------------------

(begin
  (define null-address 0)
  (define unspecified-value '<unspecified>)   ;unspecified-value

;-------------------------------------------------------------------------

; send = λεκ.κ ε

  (define (send value continuation)
    (continuation value))

  (define (unknown-variable variable)
    (define string (symbol->string variable))
    (string-append "unknown variable - " string))

;-------------------------------------------------------------------------
;---------------------------------- REP ----------------------------------
;-------------------------------------------------------------------------

  (define expression-evaluator ())

  (define (continuation-loop value)
    (lambda (environment)
      (lambda (store)
        (display value)
        (newline)
        (display ">>>")
        (define expression (read))
        ((((expression-evaluator expression continuation-loop) environment) store) environment))))

  (define (wrong message)
    (lambda (environment)
      (lambda (store)
        (lambda (recovery)
          (((continuation-loop (string-append "error: " message)) recovery) store)))))

;-------------------------------------------------------------------------
;------------------------------ constants --------------------------------
;-------------------------------------------------------------------------

  (define (constant datum)
    (if (pair? datum)
      (cons (constant (car datum)) (constant (cdr datum)))
      (if (string? datum)
        (string-append datum "")
        (if (vector? datum)
          (list->vector (constant (vector->list datum)))
          datum))))

;-------------------------------------------------------------------------
;------------------------- expression sequences --------------------------
;-------------------------------------------------------------------------

; E+⟦E+⟧ = λκ.E⟦E+ ↓ 1⟧ (λε.#E+ = 1 → (send ε κ) , E+⟦E+ † 1⟧ κ)

  (define (sequence-evaluator expressions continuation)
    (define (continuation-value value)
      (define rest-of-expressions (cdr expressions))
      (if (null? rest-of-expressions)
        (send value continuation)
        (sequence-evaluator rest-of-expressions continuation)))
    (define expression (car expressions))
    (expression-evaluator expression continuation-value))

; E*⟦E*⟧ = λκl.#E* = 0 → send l κ , E⟦E* ↓ 1⟧ (λε.E*⟦E* † 1⟧ κ (l § ⟨ε⟩))

  (define (list-evaluator expressions continuation list)
    (define (continuation-value value)
      (define extended-list (cons value list))
      (define rest-of-expressions (cdr expressions))
        (list-evaluator rest-of-expressions continuation extended-list))
    (if (null? expressions)
      (send (reverse list) continuation)
      (begin
        (define expression (car expressions))
        (expression-evaluator expression continuation-value))))

;-------------------------------------------------------------------------
;---------------------- environment/store actions ------------------------
;-------------------------------------------------------------------------
  
; initial-store = λα.α = 0 → 0

  (define (initial-store address)
    (if (= address null-address) 0))

; initial-environment = λI.∅

  (define (initial-environment variable)
    unspecified-value)

; bind-address = σ[ε/α] = λαεσ.λα'.α' = α → ε , σ α'

  (define (bind-address address value store)
    (lambda (address-prime)
      (if (= address-prime address)
        value
        (store address-prime))))

; bind-variable = ρ[α/I] = λIαρ.λI'.I = I' → α , ρ I'

  (define (bind-variable variable address environment)
    (lambda (variable-prime)
      (if (equal? variable-prime variable)
        address
        (environment variable-prime))))

; extend = λσα.σ[α/0]

  (define (extend address store)
    (bind-address null-address address store))

; locate = λασ.σ α

  (define (locate address store)
    (store address))

; lookup = λIρ.ρ I

  (define (lookup variable environment)
    (environment variable))

; new = λσ.(locate 0 σ) + 1

  (define (new store)
    (+ (locate null-address store) 1))

;-------------------------------------------------------------------------

; def = λIεκ.λρ.λσ.(λα.(λσ'.(λρ'.((send ε κ)ρ')σ')ρ[α/I])(λσ".σ"[ε/α])(extend α σ))(new σ)

  (define (def variable value continuation)
    (lambda (environment)
      (lambda (store)
        (define address (new store))
        (define updated-environment (bind-variable variable address environment))
        (define extended-store (extend address store))
        (define updated-store (bind-address address value extended-store))
        (((send value continuation) updated-environment) updated-store))))

; set = λIεκ.λρ.λσ.(λα.α = ∅ → ((wrong “unknown variable” I)ρ)σ , ((send ε κ)ρ)σ[ε/α])(lookup I ρ)

  (define (set variable value continuation)
    (lambda (environment)
      (lambda (store)
        (define address (lookup variable environment))
        (if (equal? address unspecified-value)
          (((wrong (unknown-variable variable)) environment) store)
          (begin
            (define updated-store (bind-address address value store))
            (((send value continuation) environment) updated-store))))))

; get = λIκ.λρ.λσ.(λα.α = ∅ → ((wrong “unknown variable” I)ρ)σ , ((send (locate α σ) κ)ρ)σ)(lookup I ρ)

  (define (get variable continuation)
    (lambda (environment)
      (lambda (store)
        (define address (lookup variable environment))
        (if (equal? address unspecified-value)
          (((wrong (unknown-variable variable)) environment) store)
          (begin
            (define value (locate address store))
            (((send value continuation) environment) store))))))

;-------------------------------------------------------------------------
;------------------------------ closures ---------------------------------
;-------------------------------------------------------------------------
  
; applicate = λφε*κ.φ ∈ F → φ ε* κ , wrong “procedure expected”

  (define (applicate procedure arguments continuation)
    (if (procedure? procedure) 
      (procedure arguments continuation)
      (wrong "procedure expected")))

; bind = λI*ε*κ.I* = ⟨⟩ → send ε* κ , def I*↓1 ε*↓1 (λε∅.bind I*†1 ε*†1 κ)

  (define (bind parameters arguments continuation)
    (define (continuation-bind ignore-value)
      (bind (cdr parameters) (cdr arguments) continuation))
    (if (null? parameters)
      (send arguments continuation)
      (def (car parameters) (car arguments) continuation-bind)))

; body = λE+κ.λρ.E+⟦E+⟧ (λε.λρ∅.(send ε κ)ρ)

  (define (body expressions continuation environment)
    (define (continuation-closure value)
      (lambda (ignore-environment)
        ((send value continuation) environment)))
    (sequence-evaluator expressions continuation-closure))

;-------------------------------------------------------------------------
;------------------------ expression evaluation --------------------------
;-------------------------------------------------------------------------

; E⟦(E0 E*)⟧ = λκ.E⟦E0⟧ (λε.ε ∈ F → E*⟦E*⟧ (λε*.applicate ε ε* κ) ⟨⟩ )

  (define (application-evaluator operator operands continuation)
    (define (continuation-procedure procedure)
      (define (continuation-arguments arguments)
        (applicate procedure arguments continuation))
      (list-evaluator operands continuation-arguments ()))
    (expression-evaluator operator continuation-procedure))

; E⟦(begin E+)⟧ = λκ.send E+⟦E+⟧ κ

  (define (begin-evaluator expressions continuation)
    (sequence-evaluator expressions continuation))

; E⟦K⟧ = send K⟦K⟧ κ

  (define (constant-evaluator datum continuation)
    (define constant-expression (constant datum))
    (send constant-expression continuation))

; E⟦(lambda (I*) E+)⟧ = λI*E+.λκ.λρ.(send (λε*κcall.λρcall.#I* = #ε* → 
;                                            (bind I* ε* (λε∅.body E+ κcall ρcall))ρ ,
;                                            wrong “non-matching argument list”) κ)ρ

  (define (lambda-fixed-arity-function-evaluator parameters expressions continuation)
    (lambda (environment)
      (define (procedure arguments continuation-call)
        (lambda (environment-call)
          (define (continuation-body ignore-value)
            (body expressions continuation-call environment-call))
          (if (= (length parameters) (length arguments))
            ((bind parameters arguments continuation-body) environment)
            (wrong "non-matching argument list"))))
      ((send procedure continuation) environment)))
  
; E⟦(define (I0 I*) E+)⟧ = λκ.def I0 ∅ (λε∅.E⟦(lambda (I*) E+)⟧ (λφ.set I φ κ))

  (define (define-fixed-arity-function-evaluator variable parameters expressions continuation)
    (define (continuation-define ignore-value)
      (define (continuation-closure procedure)
        (set variable procedure continuation))
      (lambda-fixed-arity-function-evaluator parameters expressions continuation-closure))
    (def variable unspecified-value continuation-define))

; E⟦(lambda (I+ . I) E+)⟧ = λI*IE+.λκ.λρ.(send (λε*κcall.λρcall.#I* <= #ε* → 
;                                                 (bind I* ε* (λε*'.def I ε*' (λε∅.body E+ κcall ρcall)))ρ ,
;                                                 wrong “non-matching argument list”) κ)ρ

  (define (lambda-variable-arity-function-evaluator parameters parameter expressions continuation)
    (lambda (environment)
      (define (procedure arguments continuation-call)
        (lambda (environment-call)
          (define (continuation-body ignore-value)
            (body expressions continuation-call environment-call))
          (define (continuation-vararg arguments)
            (def parameter arguments continuation-body))
          (if (<= (length parameters) (length arguments))
            ((bind parameters arguments continuation-vararg) environment)
            (wrong "non-matching argument list"))))
      ((send procedure continuation) environment)))

; E⟦(define I E)⟧ = λκ.def I ∅ (λε∅.E⟦E⟧ (λε.set I ε κ))

  (define (define-variable-evaluator variable expression continuation)
    (define (continuation-define ignore-value)
      (define (continuation-value value)
        (set variable value continuation))
      (expression-evaluator expression continuation-value))
    (def variable unspecified-value continuation-define))
 
; E⟦(lambda I E+)⟧ = λIE+κ.λρ.(send (λε*κcall.λρcall.def I ε* (λε∅.(body E+ κcall)ρcall)ρ) κ)ρ

  (define (lambda-list-arity-function-evaluator parameter expressions continuation)
    (lambda (environment)
      (define (procedure arguments continuation-call)
        (lambda (environment-call)
          (define (continuation-body ignore-value)
            (body expressions continuation-call environment-call))
          ((def parameter arguments continuation-body) environment)))
      ((send procedure continuation) environment)))

; E⟦(define (I0 I+ . I) E+)⟧ = λκ.def I0 ∅ (λε∅.E⟦(lambda (I+ . I) E+)⟧ (λφ.set I φ κ))

  (define (define-variable-arity-function-evaluator variable parameters parameter expressions continuation)
    (define (continuation-define ignore-value)
      (define (continuation-closure procedure)
        (set variable procedure continuation))
      (lambda-variable-arity-function-evaluator parameters parameter expressions continuation-closure))
    (def variable unspecified-value continuation-define))

; E⟦(define (I0 . I) E+)⟧ = λκ.def I0 ∅ (λε∅.E⟦(lambda I E+)⟧ (λφ.set I φ κ))

  (define (define-list-arity-function-evaluator variable parameter expressions continuation)
    (define (continuation-define ignore-value)
      (define (continuation-closure procedure)
        (set variable procedure continuation))
      (lambda-list-arity-function-evaluator parameter expressions continuation-closure))
    (def variable unspecified-value continuation-define))

; E⟦(if E0 E1)⟧ = λκ.E⟦E0⟧ (λε.λρ.ε = false → (send ∅ κ)ρ , (E⟦E1⟧ (λε.λρ∅.(send ε κ)ρ))ρ)

  (define (if-then-evaluator predicate consequent continuation)
    (define (continuation-predicate boolean)
      (lambda (environment-predicate)
        (define (continuation-clause value)
          (lambda (ignore-environment)
            ((send value continuation) environment-predicate)))
        (if boolean
          ((expression-evaluator consequent continuation-clause) environment-predicate)
          ((send unspecified-value continuation) environment-predicate))))
    (expression-evaluator predicate continuation-predicate))

; E⟦(if E0 E1 E2)⟧ = λκ.E⟦E0⟧ (λε.λρ.((ε = false → E⟦E2⟧ , E⟦E1⟧) (λε.λρ∅.(send ε κ)ρ))ρ)

  (define (if-then-else-evaluator predicate consequent alternate continuation)
    (define (continuation-predicate boolean)
      (lambda (environment-predicate)
        (define (continuation-clause value)
          (lambda (ignore-environment)
            ((send value continuation) environment-predicate)))
        (define expression (if boolean consequent alternate))
        ((expression-evaluator expression continuation-clause) environment-predicate)))
    (expression-evaluator predicate continuation-predicate))

; E⟦(quote E)⟧ = λκ.send K⟦E⟧ κ

  (define (quote-evaluator quoted-expression continuation)
    (define constant-expression (constant quoted-expression))
    (send constant-expression continuation))

; E⟦(set! I E)⟧ = λκ.E⟦E⟧ (λε.set I ε κ)

  (define (set!-evaluator variable expression continuation)
    (define (continuation-value value)
      (set variable value continuation))
    (expression-evaluator expression continuation-value))

; E⟦I⟧ = λκ.get I κ

  (define (variable-evaluator variable continuation)
    (get variable continuation))

;-------------------------------------------------------------------------
;-------------------------- expression parsing ---------------------------
;-------------------------------------------------------------------------

  (define (parse-expression-sequence operands)
    (define (proper-list expression)
      (if (pair? expression)
        (proper-list (cdr expression))
        (null? expression)))
    (if (null? operands)
      (wrong "empty expression sequence")
      (if (proper-list operands)
        operands
        (wrong "improper expression sequence"))))

  (define (parse-pair operands process)
    (if (pair? operands)
      (process (car operands) (cdr operands))
      (wrong "pair required"))) 

  (define (parse-parameter-pattern parameter-pattern process)
    (define (parse-pattern parameter-pattern parameter-list)
      (define (parse-parameter parameter residue)
        (if (symbol? parameter)
          (parse-pattern residue (cons parameter parameter-list))
          (wrong "parameter required"))) 
      (if (null? parameter-pattern)
        (process (reverse parameter-list) ())
        (if (symbol? parameter-pattern)
          (process (reverse parameter-list) parameter-pattern)
          (parse-pair parameter-pattern parse-parameter))))
    (parse-pattern parameter-pattern ()))

  (define (parse-begin operands continuation)
    (define expression-sequence (parse-expression-sequence operands))
    (begin-evaluator expression-sequence continuation))

  (define (parse-define operands continuation)
    (define (parse-define-binding pattern body)
      (define (parse-define-variable expression residue)
        (if (null? residue)
          (define-variable-evaluator pattern expression continuation)
          (wrong "single expression required in variable <define>")))
      (define (parse-define-pattern variable parameter-pattern)
        (define (parse-define-function parameters parameter)
          (define expressions (parse-expression-sequence body))
          (if (symbol? parameter)
            (if (null? parameters)
              (define-list-arity-function-evaluator variable parameter expressions continuation)
              (define-variable-arity-function-evaluator variable parameters parameter expressions continuation))
            (define-fixed-arity-function-evaluator variable parameters expressions continuation))) 
        (if (symbol? variable)
          (parse-parameter-pattern parameter-pattern parse-define-function)
          (wrong "variable required in function <define>")))
      (if (symbol? pattern)
        (parse-pair body parse-define-variable)
        (parse-pair pattern parse-define-pattern)))
    (parse-pair operands parse-define-binding))

  (define (parse-if operands continuation)
    (define (parse-predicate predicate clauses)
      (define (parse-clauses consequent residue)
        (define (parse-alternative alternative residue)
          (if (null? residue)
            (if-then-else-evaluator predicate consequent alternative continuation)
            (wrong "too many clauses in <if>")))
        (if (null? residue)
          (if-then-evaluator predicate consequent continuation)
          (parse-pair residue parse-alternative)))
      (parse-pair clauses parse-clauses))
    (parse-pair operands parse-predicate))

  (define (parse-lambda operands continuation)
    (define (parse-lambda-pattern parameter-pattern body)
      (define (parse-lambda-function parameters parameter)
        (define expressions (parse-expression-sequence body))
        (if (symbol? parameter)
          (if (null? parameters)
            (lambda-list-arity-function-evaluator parameter expressions continuation)
            (lambda-variable-arity-function-evaluator parameters parameter expressions continuation))
          (lambda-fixed-arity-function-evaluator parameters expressions continuation))) 
      (parse-parameter-pattern parameter-pattern parse-lambda-function))
    (parse-pair operands parse-lambda-pattern))

  (define (parse-quote operands continuation)
    (define (parse-quote-expression quote-expression residue)
      (if (null? residue)
        (quote-evaluator (car operands) continuation)
        (wrong "single expression required in quotation")))
    (parse-pair operands parse-quote-expression))

  (define (parse-set! operands continuation)
    (define (parse-set!-binding variable body)
      (define (parse-set!-variable expression residue)
        (if (null? residue)
          (set!-evaluator variable expression continuation)
          (wrong "single expression required in assignment")))
      (if (symbol? variable)
        (parse-pair body parse-set!-variable)
        (wrong "variable required in assignment")))
    (parse-pair operands parse-set!-binding))

;-------------------------------------------------------------------------
;------------------------ expression dispatcher --------------------------
;-------------------------------------------------------------------------

  (set! expression-evaluator 
    (lambda (expression continuation)
      (if (pair? expression)
        (begin
          (define operator (car expression))
          (define operands (cdr expression))
          (if (equal? operator 'begin) 
            (parse-begin operands continuation)
            (if (equal? operator 'define)  
              (parse-define operands continuation)
              (if (equal? operator 'if)  
                (parse-if operands continuation) 
                (if (equal? operator 'lambda)  
                  (parse-lambda operands continuation)
                  (if (equal? operator 'quote)  
                    (parse-quote operands continuation)
                    (if (equal? operator 'set!)  
                      (parse-set! operands continuation)
                      (application-evaluator operator operands continuation)))))))) 
        (if (symbol? expression)
          (variable-evaluator expression continuation)
          (constant-evaluator expression continuation)))))

;-------------------------------------------------------------------------
;------------------------- primitive functions ---------------------------
;-------------------------------------------------------------------------

  (define (wrap primitive)
    (lambda (arguments continuation)
      (define value (apply primitive arguments))
      (send value continuation)))

  (define (ds-apply arguments continuation)
    (define (parse-procedure procedure residue)
      (define (parse-argument-list argument-list residue)
        (if (null? residue)
          (applicate procedure argument-list continuation) 
          (wrong "two arguments required in apply")))
      (parse-pair residue parse-argument-list))
    (parse-pair arguments parse-procedure))

  (define (ds-call-with-current-continuation arguments continuation)
    (lambda (environment)
      (define (current-continuation arguments ignore-continuation)
        (define (parse-value value residue)
          (if (null? residue)
            ((send value continuation) environment)
            (wrong "one argument required in current continuation")))
        (parse-pair arguments parse-value))
      (define (parse-procedure procedure residue)
        (if (null? residue)
          ((applicate procedure (list current-continuation) continuation) environment) 
          (wrong "one argument required in call-with-current-continuation")))
      (parse-pair arguments parse-procedure)))

  (define (ds-eval arguments continuation)
    (define (parse-expression expression residue)
      (if (null? residue)
        (expression-evaluator expression continuation))
        (wrong "one argument required in eval"))
    (parse-pair arguments parse-expression))

  (define (ds-for-each arguments continuation)
    (wrong "not yet implemented: for-each"))

  (define (ds-force arguments continuation)
    (wrong "not yet implemented: force"))

  (define (ds-load arguments continuation)
    (define (parse-string string residue)
      (if (null? residue)
        (if (string? string)
          (begin
            (define port (open-input-file string))
            (define expression (read port))
            (close-input-port port)
            (expression-evaluator expression continuation))
          (wrong "string argument required in load"))
        (wrong "one argument required in load")))
    (parse-pair arguments parse-string))

  (define (ds-map arguments continuation)
    (wrong "not yet implemented: map"))

  (define (ds-cons-stream arguments continuation)
    (wrong "not yet implemented: cons-stream"))

  (define (ds-error arguments continuation)
    (define (parse-message message residue)
      (if (null? residue)
        (wrong message)
        (wrong "one argument required in error")))
    (parse-pair arguments parse-message))

  (define primitives (list (cons 'circularity-level 0)
                           (cons 'false #f)
                           (cons 'true #t)
                           (cons '- (wrap -))
                           (cons '* (wrap *))
                           (cons '/ (wrap /))
                           (cons '+ (wrap +))
                           (cons '< (wrap <))
                           (cons '<= (wrap <=))
                           (cons '= (wrap =))
                           (cons '> (wrap >))
                           (cons '>= (wrap >=))
                           (cons 'abs (wrap abs))
                           (cons 'acos (wrap acos))
                           (cons 'append (wrap append))
                           (cons 'asin (wrap asin))
                           (cons 'assoc (wrap assoc))
                           (cons 'assq (wrap assq))
                           (cons 'atan (wrap atan))
                           (cons 'boolean? (wrap boolean?))
                           (cons 'caaaar (wrap caaaar))
                           (cons 'caaadr (wrap caaadr))
                           (cons 'caaar (wrap caaar))
                           (cons 'caadar (wrap caadar))
                           (cons 'caaddr (wrap caaddr))
                           (cons 'caadr (wrap caadr))
                           (cons 'caar (wrap caar))
                           (cons 'cadaar (wrap cadaar))
                           (cons 'cadadr (wrap cadadr))
                           (cons 'cadar (wrap cadar))
                           (cons 'caddar (wrap caddar))
                           (cons 'cadddr (wrap cadddr))
                           (cons 'caddr (wrap caddr))
                           (cons 'cadr (wrap cadr))
                           (cons 'car (wrap car))
                           (cons 'cdaaar (wrap cdaaar))
                           (cons 'cdaadr (wrap cdaadr))
                           (cons 'cdaar (wrap cdaar))
                           (cons 'cdadar (wrap cdadar))
                           (cons 'cdaddr (wrap cdaddr))
                           (cons 'cdadr (wrap cdadr))
                           (cons 'cdar (wrap cdar))
                           (cons 'cddaar (wrap cddaar))
                           (cons 'cddadr (wrap cddadr))
                           (cons 'cddar (wrap cddar))
                           (cons 'cdddar (wrap cdddar))
                           (cons 'cddddr (wrap cddddr))
                           (cons 'cdddr (wrap cdddr))
                           (cons 'cddr (wrap cddr))
                           (cons 'cdr (wrap cdr))
                           (cons 'char->integer (wrap char->integer))
                           (cons 'char? (wrap char?))
                           (cons 'char<? (wrap char<?))
                           (cons 'char<=? (wrap char<=?))
                           (cons 'char=? (wrap char=?))
                           (cons 'char>? (wrap char>?))
                           (cons 'char>=? (wrap char>=?))
                           (cons 'close-input-port (wrap close-input-port))
                           (cons 'close-output-port (wrap close-output-port))
                           (cons 'cons (wrap cons))
                           (cons 'cos (wrap cos))
                           (cons 'display (wrap display))
                           (cons 'eof-object? (wrap eof-object?))
                           (cons 'eq? (wrap eq?))
                           (cons 'equal? (wrap equal?))
                           (cons 'even? (wrap even?))
                           (cons 'exp (wrap exp))
                           (cons 'expt (wrap expt))
                           (cons 'input-port? (wrap input-port?))
                           (cons 'integer? (wrap integer?))
                           (cons 'length (wrap length))
                           (cons 'list (wrap list))
                           (cons 'list->vector (wrap list->vector))
                           (cons 'log (wrap log))
                           (cons 'make-string (wrap make-string))
                           (cons 'make-vector (wrap make-vector))
                           (cons 'max (wrap max))
                           (cons 'member (wrap member))
                           (cons 'memq (wrap memq))
                           (cons 'min (wrap min))
                           (cons 'modulo (wrap modulo))
                           (cons 'negative? (wrap negative?))
                           (cons 'newline (wrap newline))
                           (cons 'not (wrap not))
                           (cons 'null? (wrap null?))
                           (cons 'number->string (wrap number->string))
                           (cons 'number? (wrap number?))
                           (cons 'odd? (wrap odd?))
                           (cons 'open-input-file (wrap open-input-file))
                           (cons 'open-output-file (wrap open-output-file))
                           (cons 'output-port? (wrap output-port?))
                           (cons 'pair? (wrap pair?))
                           (cons 'positive? (wrap positive?))
                           (cons 'procedure? (wrap procedure?))
                           (cons 'quotient (wrap quotient))
                           (cons 'read (wrap read))
                           (cons 'read-char (wrap read-char))
                           (cons 'real? (wrap real?))
                           (cons 'remainder (wrap remainder))
                           (cons 'reverse (wrap reverse))
                           (cons 'set-car! (wrap set-car!))
                           (cons 'set-cdr! (wrap set-cdr!))
                           (cons 'sin (wrap sin))
                           (cons 'sqrt (wrap sqrt))
                           (cons 'string->number (wrap string->number))
                           (cons 'string->symbol (wrap string->symbol))
                           (cons 'string-append (wrap string-append))
                           (cons 'string-length (wrap string-length))
                           (cons 'string-ref (wrap string-ref))
                           (cons 'string-set! (wrap string-set!))
                           (cons 'string? (wrap string?))
                           (cons 'string<? (wrap string<?))
                           (cons 'string<=? (wrap string<=?))
                           (cons 'string=? (wrap string=?))
                           (cons 'string>? (wrap string>?))
                           (cons 'string>=? (wrap string>=?))
                           (cons 'substring (wrap substring))
                           (cons 'symbol->string (wrap symbol->string))
                           (cons 'symbol? (wrap symbol?))
                           (cons 'tan (wrap tan))
                           (cons 'vector (wrap vector))
                           (cons 'vector->list (wrap vector->list))
                           (cons 'vector-length (wrap vector-length))
                           (cons 'vector-ref (wrap vector-ref))
                           (cons 'vector-set! (wrap vector-set!))
                           (cons 'vector? (wrap vector?))
                           (cons 'write (wrap write))
                           (cons 'write-char (wrap write-char))
                           (cons 'zero? (wrap zero?))
                         ; (cons 'stream-null? (wrap stream-null?))
                         ; (cons 'the-empty-stream the-empty-stream)
                         ; (cons 'stream-car (wrap stream-car))
                         ; (cons 'stream-cdr (wrap stream-cdr))
                         ; (cons 'clock (wrap clock))
                         ; (cons 'collect (wrap collect))
                         ; (cons 'pretty (wrap pretty))
                         ; (cons 'random (wrap random))
                           (cons 'apply ds-apply)
                           (cons 'call-with-current-continuation ds-call-with-current-continuation)
                           (cons 'eval ds-eval)
                           (cons 'for-each ds-for-each)
                         ; (cons 'force ds-force)
                           (cons 'load ds-load)
                           (cons 'map ds-map)
                         ; (cons 'cons-stream ds-cons-stream)
                           (cons 'error ds-error)))
 
;-------------------------------------------------------------------------
;---------------------------- initialisation -----------------------------
;-------------------------------------------------------------------------

  (define (initialize primitives)
    (if (null? primitives)
      (continuation-loop "Denotational Semantics Slip")
      (begin
        (define (continuation-define ignore-value)
          (initialize (cdr primitives)))
        (define variable (caar primitives))
        (define value (cdar primitives))  
        (def variable value continuation-define))))

  (((initialize primitives) initial-environment) initial-store))